@@ -1,6 +1,25 @@
 #include "perl.h"
 #include <stdlib.h>
 #include <string.h>
+#include <ctype.h>
+
+#ifndef Uri_Size
+#define Uri_Size_scheme 32
+#define Uri_Size_path   1024
+#define Uri_Size_query  2048
+#define Uri_Size_frag   64
+#define Uri_Size_usr    64
+#define Uri_Size_pwd    64
+#define Uri_Size_host   256
+#define Uri_Size_port   8
+// enough to fit all pieces + 3 chars for separators (2 colons + @)
+#define Uri_Size_auth   (3 + Uri_Size_usr + Uri_Size_pwd + Uri_Size_host + Uri_Size_port)
+#define Uri_Size(member) (Uri_Size_##member)
+#endif
+
+#ifndef Uri_Encode_Set_Nolen
+#define Uri_Encode_Set_Nolen(uri_obj, member, value, allow, allow_len) uri_encode(value, min(strlen(value), Uri_Size(member)), Uri_Mem(uri_obj, member), allow, allow_len)
+#endif
 
 /*
  * Allocate memory with Newx if it's
@@ -221,80 +240,103 @@ SV* decode(SV* in) {
 /*
  * Internal API
  */
+typedef char uri_scheme_t [Uri_Size_scheme + 1];
+typedef char uri_path_t   [Uri_Size_path + 1];
+typedef char uri_query_t  [Uri_Size_query + 1];
+typedef char uri_frag_t   [Uri_Size_frag + 1];
+typedef char uri_usr_t    [Uri_Size_usr + 1];
+typedef char uri_pwd_t    [Uri_Size_pwd + 1];
+typedef char uri_host_t   [Uri_Size_host + 1];
+typedef char uri_port_t   [Uri_Size_port + 1];
 
 typedef struct {
-  char scheme[16];
-  char auth[264];
-  char path[256];
-  char query[1024];
-  char frag[32];
-
-  char usr[64];
-  char pwd[64];
-  char host[128];
-  char port[8];
+  uri_scheme_t scheme;
+  uri_path_t   path;
+  uri_query_t  query;
+  uri_frag_t   frag;
+  uri_usr_t    usr;
+  uri_pwd_t    pwd;
+  uri_host_t   host;
+  uri_port_t   port;
 } uri_t;
 
+inline
+size_t min(size_t a, size_t b) {
+  return a <= b ? a : b;
+}
+
 /*
  * Clearers
- *   -note that these do not do other related cleanup (e.g. clearing auth triggering
- *    the clearing of usr/pwd/host/port)
  */
-void clear_scheme(SV* uri_obj) { memset(&((Uri(uri_obj))->scheme), '\0', 16);   }
-void clear_auth(SV* uri_obj)   { memset(&((Uri(uri_obj))->auth),   '\0', 264);  }
-void clear_path(SV* uri_obj)   { memset(&((Uri(uri_obj))->path),   '\0', 256);  }
-void clear_query(SV* uri_obj)  { memset(&((Uri(uri_obj))->query),  '\0', 1024); }
-void clear_frag(SV* uri_obj)   { memset(&((Uri(uri_obj))->frag),   '\0', 32);   }
-void clear_usr(SV* uri_obj)    { memset(&((Uri(uri_obj))->usr),    '\0', 64);   }
-void clear_pwd(SV* uri_obj)    { memset(&((Uri(uri_obj))->pwd),    '\0', 64);   }
-void clear_host(SV* uri_obj)   { memset(&((Uri(uri_obj))->host),   '\0', 128);  }
-void clear_port(SV* uri_obj)   { memset(&((Uri(uri_obj))->port),   '\0', 8);    }
+void clear_scheme(SV* uri_obj) { memset(&((Uri(uri_obj))->scheme), '\0', sizeof(uri_scheme_t)); }
+void clear_path(SV* uri_obj)   { memset(&((Uri(uri_obj))->path),   '\0', sizeof(uri_path_t));   }
+void clear_query(SV* uri_obj)  { memset(&((Uri(uri_obj))->query),  '\0', sizeof(uri_query_t));  }
+void clear_frag(SV* uri_obj)   { memset(&((Uri(uri_obj))->frag),   '\0', sizeof(uri_frag_t));   }
+void clear_usr(SV* uri_obj)    { memset(&((Uri(uri_obj))->usr),    '\0', sizeof(uri_usr_t));    }
+void clear_pwd(SV* uri_obj)    { memset(&((Uri(uri_obj))->pwd),    '\0', sizeof(uri_pwd_t));    }
+void clear_host(SV* uri_obj)   { memset(&((Uri(uri_obj))->host),   '\0', sizeof(uri_host_t));   }
+void clear_port(SV* uri_obj)   { memset(&((Uri(uri_obj))->port),   '\0', sizeof(uri_port_t));   }
+
+void clear_auth(SV* uri_obj) {
+  clear_usr(uri_obj);
+  clear_pwd(uri_obj);
+  clear_host(uri_obj);
+  clear_port(uri_obj);
+}
 
 /*
- * Scans the authorization portion of the Uri string. This must only be called
- * *after* the 'auth' member has been populated (eg, by uri_scan).
+ * Scans the authorization portion of the Uri string
  */
-void uri_scan_auth (uri_t* uri) {
-  size_t len  = strlen((char*) uri->auth);
+void uri_scan_auth(uri_t* uri, const char* auth, const size_t len) {
   size_t idx  = 0;
   size_t brk1 = 0;
   size_t brk2 = 0;
+  size_t i;
 
-  memset(&uri->usr,  '\0', 64);
-  memset(&uri->pwd,  '\0', 64);
-  memset(&uri->host, '\0', 128);
-  memset(&uri->port, '\0', 8);
+  memset(&uri->usr,  '\0', sizeof(uri_usr_t));
+  memset(&uri->pwd,  '\0', sizeof(uri_pwd_t));
+  memset(&uri->host, '\0', sizeof(uri_host_t));
+  memset(&uri->port, '\0', sizeof(uri_port_t));
 
   if (len > 0) {
     // Credentials
-    brk1 = strcspn(&uri->auth[idx], "@");
+    brk1 = strcspn(&auth[idx], "@");
 
     if (brk1 > 0 && brk1 != len) {
-      brk2 = strcspn(&uri->auth[idx], ":");
+      brk2 = strcspn(&auth[idx], ":");
 
       if (brk2 > 0 && brk2 < brk1) {
-        strncpy(uri->usr, &uri->auth[idx], brk2);
+        strncpy(uri->usr, &auth[idx], min(brk2, Uri_Size_usr));
         idx += brk2 + 1;
 
-        strncpy(uri->pwd, &uri->auth[idx], brk1 - brk2 - 1);
+        strncpy(uri->pwd, &auth[idx], min(brk1 - brk2 - 1, Uri_Size_pwd));
         idx += brk1 - brk2;
       }
       else {
-        strncpy(uri->usr, &uri->auth[idx], brk1);
+        strncpy(uri->usr, &auth[idx], min(brk1, Uri_Size_usr));
         idx += brk1 + 1;
       }
     }
 
     // Location
-    brk1 = strcspn(&uri->auth[idx], ":");
+    brk1 = strcspn(&auth[idx], ":");
 
     if (brk1 > 0 && brk1 != (len - idx)) {
-      strncpy(uri->host, &uri->auth[idx], brk1);
+      strncpy(uri->host, &auth[idx], min(brk1, Uri_Size_host));
       idx += brk1 + 1;
-      strncpy(uri->port, &uri->auth[idx], len - idx);
+
+      for (i = 0; i < (len - idx) && i < Uri_Size_port; ++i) {
+        if (!isdigit(auth[i + idx])) {
+          memset(&uri->port, '\0', Uri_Size_port + 1);
+          break;
+        }
+        else {
+          uri->port[i] = auth[i + idx];
+        }
+      }
     }
     else {
-      strncpy(uri->host, &uri->auth[idx], len - idx);
+      strncpy(uri->host, &auth[idx], min(len - idx, Uri_Size_host));
     }
   }
 }
@@ -309,89 +351,48 @@ void uri_scan(uri_t* uri, const char* src, size_t len) {
   // Scheme
   brk = strcspn(&src[idx], ":/@?#");
   if (brk > 0 && strncmp(&src[idx + brk], "://", 3) == 0) {
-    strncpy(uri->scheme, &src[idx], brk);
+    strncpy(uri->scheme, &src[idx], min(brk, Uri_Size_scheme));
     uri->scheme[brk] = '\0';
     idx += brk + 3;
 
     // Authority
     brk = strcspn(&src[idx], "/?#");
     if (brk > 0) {
-      strncpy(uri->auth, &src[idx], brk);
-      uri->auth[brk] = '\0';
+      uri_scan_auth(uri, &src[idx], brk);
       idx += brk;
     }
   }
 
-  // Path
+  // path
   brk = strcspn(&src[idx], "?#");
   if (brk > 0) {
-    strncpy(uri->path, &src[idx], brk);
+    strncpy(uri->path, &src[idx], min(brk, Uri_Size_path));
     uri->path[brk] = '\0';
     idx += brk;
   }
 
-  // Query
+  // query
   if (src[idx] == '?') {
     ++idx; // skip past ?
     brk = strcspn(&src[idx], "#");
     if (brk > 0) {
-      strncpy(uri->query, &src[idx], brk);
+      strncpy(uri->query, &src[idx], min(brk, Uri_Size_query));
       uri->query[brk] = '\0';
       idx += brk;
     }
   }
 
-  // Fragment
+  // fragment
   if (src[idx] == '#') {
     ++idx; // skip past #
     brk = len - idx;
     if (brk > 0) {
-      strncpy(uri->frag, &src[idx], brk);
+      strncpy(uri->frag, &src[idx], min(brk, Uri_Size_frag));
       uri->frag[brk] = '\0';
     }
   }
 }
 
-/*
- * Rebuilds the authority string: username:password@hostname:portnumber
- */
-void uri_build_auth(uri_t* uri) {
-  size_t len = 0;
-  int idx = 0;
-
-  memset(&uri->auth, '\0', 264);
-
-  if (uri->usr[0] != '\0') {
-    len = strlen((char*) &uri->usr);
-    strncpy(&uri->auth[idx], (char*) &uri->usr, len);
-    idx += len;
-
-    if (uri->pwd[0] != '\0') {
-      len = strlen((char*) &uri->pwd);
-      uri->auth[idx++] = ':';
-      strncpy(&uri->auth[idx], (char*) &uri->pwd, len);
-      idx += len;
-    }
-
-    uri->auth[idx++] = '@';
-  }
-
-  if (uri->host[0] != '\0') {
-    len = strlen((char*) &uri->host);
-    strncpy(&uri->auth[idx], (char*) &uri->host, len);
-    idx += len;
-
-    if (uri->port[0] != '\0') {
-      len = strlen((char*) &uri->port);
-      uri->auth[idx++] = ':';
-      strncpy(&uri->auth[idx], (char*) &uri->port, len);
-      idx += len;
-    }
-  }
-
-  uri->auth[idx++] = '\0';
-}
-
 /*
  * Perl API
  */
@@ -400,37 +401,64 @@ void uri_build_auth(uri_t* uri) {
  * Getters
  */
 const char* get_scheme(SV* uri_obj) { return Uri_Mem(uri_obj, scheme); }
-const char* get_auth(SV* uri_obj)   { return Uri_Mem(uri_obj, auth); }
-const char* get_path(SV* uri_obj)   { return Uri_Mem(uri_obj, path); }
-const char* get_query(SV* uri_obj)  { return Uri_Mem(uri_obj, query); }
-const char* get_frag(SV* uri_obj)   { return Uri_Mem(uri_obj, frag); }
-const char* get_usr(SV* uri_obj)    { return Uri_Mem(uri_obj, usr); }
-const char* get_pwd(SV* uri_obj)    { return Uri_Mem(uri_obj, pwd); }
-const char* get_host(SV* uri_obj)   { return Uri_Mem(uri_obj, host); }
-const char* get_port(SV* uri_obj)   { return Uri_Mem(uri_obj, port); }
+const char* get_path(SV* uri_obj)   { return Uri_Mem(uri_obj, path);   }
+const char* get_query(SV* uri_obj)  { return Uri_Mem(uri_obj, query);  }
+const char* get_frag(SV* uri_obj)   { return Uri_Mem(uri_obj, frag);   }
+const char* get_usr(SV* uri_obj)    { return Uri_Mem(uri_obj, usr);    }
+const char* get_pwd(SV* uri_obj)    { return Uri_Mem(uri_obj, pwd);    }
+const char* get_host(SV* uri_obj)   { return Uri_Mem(uri_obj, host);   }
+const char* get_port(SV* uri_obj)   { return Uri_Mem(uri_obj, port);   }
+
+SV* get_auth(SV* uri_obj) {
+  uri_t* uri = Uri(uri_obj);
+  SV* out = newSVpv("", 0);
+
+  if (uri->usr[0] != '\0') {
+    if (uri->pwd[0] != '\0') {
+      sv_catpvf(out, "%s:%s@", uri->usr, uri->pwd);
+    } else {
+      sv_catpvf(out, "%s@", uri->usr);
+    }
+  }
+
+  if (uri->host[0] != '\0') {
+    if (uri->port[0] != '\0') {
+      sv_catpvf(out, "%s:%s", uri->host, uri->port);
+    } else {
+      sv_catpv(out, uri->host);
+    }
+  }
+
+  return out;
+}
 
 SV* query_hash(SV* uri) {
   const char* src = Uri_Mem(uri, query);
   size_t klen, vlen, idx;
   HV*  out = newHV();
-  AV*  arr;
-  SV** ref;
-  SV*  tmp;
+  AV*  arr = NULL;
+  SV** ref = NULL;
+  SV*  tmp = NULL;
 
   while (src != NULL && src[0] != '\0') {
     idx = strcspn(src, "=");
+    if (idx == 0) break;
+
     char key[idx + 1];
     klen = uri_decode(src, idx, key);
 
     src = strstr(src, "=");
-    src += 1;
 
-    idx = strcspn(src, "&");
-    char val[idx + 1];
-    vlen = uri_decode(src, idx, val);
+    if (src != NULL) {
+      src += 1;
 
-    tmp = newSVpv(val, vlen);
-    SvUTF8_on(tmp);
+      idx = strcspn(src, "&");
+      char val[idx + 1];
+      vlen = uri_decode(src, idx, val);
+
+      tmp = newSVpv(val, vlen);
+      SvUTF8_on(tmp);
+    }
 
     if (!hv_exists(out, key, klen)) {
       arr = newAV();
@@ -442,10 +470,15 @@ SV* query_hash(SV* uri) {
       arr = (AV*) SvRV(*ref);
     }
 
-    av_push(arr, tmp);
+    if (tmp != NULL) {
+      av_push(arr, tmp);
+    }
+
+    if (src == NULL) break;
 
     src = strstr(src, "&");
     if (src == NULL) break;
+
     ++src;
   }
 
@@ -529,53 +562,62 @@ SV* get_param(SV* uri, const char* key) {
  * Setters
  */
 
-const char* set_scheme(SV* uri_obj, const char* value, int no_triggers) {
-  uri_encode(value, 0, Uri_Mem(uri_obj, scheme), "", 0);
+const char* set_scheme(SV* uri_obj, const char* value) {
+  Uri_Encode_Set_Nolen(uri_obj, scheme, value, "", 0);
   return Uri_Mem(uri_obj, scheme);
 }
 
-const char* set_auth(SV* uri_obj, const char* value, int no_triggers) {
-  uri_encode(value, 0, Uri_Mem(uri_obj, auth), ":@", 2);
-  if (!no_triggers) uri_scan_auth(Uri(uri_obj));
-  return Uri_Mem(uri_obj, auth);
+SV* set_auth(SV* uri_obj, const char* value) {
+  char auth[Uri_Size_auth];
+  size_t len = uri_encode(value, strlen(value), &auth, ":@", 2);
+  uri_scan_auth(Uri(uri_obj), auth, len);
+  return newSVpv(auth, len);
 }
 
-const char* set_path(SV* uri_obj, const char* value, int no_triggers) {
-  uri_encode(value, 0, Uri_Mem(uri_obj, path), "/", 1);
+const char* set_path(SV* uri_obj, const char* value) {
+  Uri_Encode_Set_Nolen(uri_obj, path, value, "/", 1);
   return Uri_Mem(uri_obj, path);
 }
 
-const char* set_query(SV* uri_obj, const char* value, int no_triggers) {
-  strncpy(Uri_Mem(uri_obj, query), value, strlen(value) + 1);
+const char* set_query(SV* uri_obj, const char* value) {
+  strncpy(Uri_Mem(uri_obj, query), value, min(strlen(value) + 1, Uri_Size_query));
   return value;
 }
 
-const char* set_frag(SV* uri_obj, const char* value, int no_triggers) {
-  uri_encode(value, 0, Uri_Mem(uri_obj, frag), "", 0);
+const char* set_frag(SV* uri_obj, const char* value) {
+  Uri_Encode_Set_Nolen(uri_obj, frag, value, "", 0);
   return Uri_Mem(uri_obj, frag);
 }
 
-const char* set_usr(SV* uri_obj, const char* value, int no_triggers) {
-  uri_encode(value, 0, Uri_Mem(uri_obj, usr), "", 0);
-  if (!no_triggers) uri_build_auth(Uri(uri_obj));
+const char* set_usr(SV* uri_obj, const char* value) {
+  Uri_Encode_Set_Nolen(uri_obj, usr, value, "", 0);
   return Uri_Mem(uri_obj, usr);
 }
 
-const char* set_pwd(SV* uri_obj, const char* value, int no_triggers) {
-  uri_encode(value, 0, Uri_Mem(uri_obj, pwd), "", 0);
-  if (!no_triggers) uri_build_auth(Uri(uri_obj));
+const char* set_pwd(SV* uri_obj, const char* value) {
+  Uri_Encode_Set_Nolen(uri_obj, pwd, value, "", 0);
   return Uri_Mem(uri_obj, pwd);
 }
 
-const char* set_host(SV* uri_obj, const char* value, int no_triggers) {
-  uri_encode(value, 0, Uri_Mem(uri_obj, host), "", 0);
-  if (!no_triggers) uri_build_auth(Uri(uri_obj));
+const char* set_host(SV* uri_obj, const char* value) {
+  Uri_Encode_Set_Nolen(uri_obj, host, value, "", 0);
   return Uri_Mem(uri_obj, host);
 }
 
-const char* set_port(SV* uri_obj, const char* value, int no_triggers) {
-  uri_encode(value, 0, Uri_Mem(uri_obj, port), "", 0);
-  if (!no_triggers) uri_build_auth(Uri(uri_obj));
+const char* set_port(SV* uri_obj, const char* value) {
+  size_t len = min(strlen(value), Uri_Size_port);
+  size_t i;
+
+  for (i = 0; i < len; ++i) {
+    if (isdigit(value[i])) {
+      Uri_Mem(uri_obj, port)[i] = value[i];
+    }
+    else {
+      clear_port(uri_obj);
+      break;
+    }
+  }
+
   return Uri_Mem(uri_obj, port);
 }
 
@@ -655,7 +697,7 @@ SV* to_string(SV* uri_obj) {
 
   sv_catpv(out, uri->scheme);
   sv_catpv(out, "://");
-  sv_catpv(out, uri->auth);
+  sv_catsv(out, sv_2mortal(get_auth(uri_obj)));
   sv_catpv(out, uri->path);
 
   if (uri->query[0] != '\0') {
@@ -673,7 +715,7 @@ SV* to_string(SV* uri_obj) {
 
 void explain(SV* uri_obj) {
   printf("scheme: %s\n",  Uri_Mem(uri_obj, scheme));
-  printf("auth: %s\n",    Uri_Mem(uri_obj, auth));
+  printf("auth:\n");
   printf("  -usr: %s\n",  Uri_Mem(uri_obj, usr));
   printf("  -pwd: %s\n",  Uri_Mem(uri_obj, pwd));
   printf("  -host: %s\n", Uri_Mem(uri_obj, host));
@@ -700,7 +742,6 @@ SV* new(const char* class, SV* uri_str) {
 
   src = SvPV_const(uri_str, len);
   uri_scan(uri, src, len);
-  uri_scan_auth(uri);
 
   return obj_ref;
 }
@@ -742,7 +783,7 @@ void uri_split(SV* uri) {
     Inline_Stack_Push(&PL_sv_undef);
   }
 
-  // Path
+  // path
   brk = strcspn(&src[idx], "?#");
   if (brk > 0) {
     Inline_Stack_Push(sv_2mortal(newSVpv(&src[idx], brk)));
@@ -751,7 +792,7 @@ void uri_split(SV* uri) {
     Inline_Stack_Push(sv_2mortal(newSVpv("",0)));
   }
 
-  // Query
+  // query
   if (src[idx] == '?') {
     ++idx; // skip past ?
     brk = strcspn(&src[idx], "#");
@@ -765,7 +806,7 @@ void uri_split(SV* uri) {
     Inline_Stack_Push(&PL_sv_undef);
   }
 
-  // Fragment
+  // fragment
   if (src[idx] == '#') {
     ++idx; // skip past #
     brk = len - idx;
